module Splot
  class Command
    attr :include_path, :line_no, :file, :preloader

    def initialize(params = {})
      @file         = params.fetch(:file).to_str
      @include_path = params[:include_path]
      @line_no      = Integer(params[:line]) if params.has_key? :line
      @preloader    = params[:preloader].to_str if params.has_key? :preloader
    end

    def self.for(file)
      ObjectSpace.each_object(Class).select { |klass| klass < self }.each do |subklass|
        return subklass if subklass.match? file
      end
      self
    end

    def self.match?(file)
      false
    end

    def command
      [
        preloader,    # "spring"
        runner,       # "rake test"
        include_args, # "-Itest"
        file_arg,     # "TEST=\"/path/to/test.rb\""
        line_no_arg,  # "TESTOPTS=\"-n test_foo\""
      ].compact * ' '
    end

    def file_arg
      file
    end

    def include_args
      include_path ? "-I#{include_path}" : nil
    end

    def line_no_arg
      line_no ? "-l #{line_no}" : nil
    end

    def runner
      raise RunnerNotFoundError, "cannot determine runner for #{file.inspect}"
    end
  end

  class CucumberCommand < Command
    def self.match?(file)
      file.match %r{\.feature\Z}
    end

    def include_args
      return super if include_path.nil?
      raise ArgumentError, "Cucumber does not support include arguments"
    end

    def runner
      'cucumber'
    end
  end

  class RSpecCommand < Command
    def self.match?(file)
      file.match /\_spec\.rb\Z/
    end

    def runner
      'rspec'
    end
  end

  class TestUnitCommand < Command

    ACTIVE_SUPPORT_TEST_CASE_MATCHER = %r{
      \A\s*test\s+              #   test
      (?:                       #
         "(?<double_quote>.*)"  #        "it should work"
         |                      # OR
         '(?<single_quote>.*)'  #        'it should work'
      )                         #
      \s+do\s*\Z                #                         do\n
    }x

    TEST_UNIT_MATCHER = %r{
      \A\s*def\s+               #  def
      (?<test_name>             #
        test_                   #      test_
        [a-zA-Z_]+              #           it_should_work
        [\?!=]?                 #                         ?
      )                         #
      \s*\Z                     #                          \n
    }x

    def initialize(params = {})
      @runner = params.fetch(:runner) { "rake test" }
      super
    end

    def self.match?(file)
      file.match /\_test\.rb/
    end

    def file_arg
      if rake?
        "TEST=#{file}"
      else
        super
      end
    end

    def rake?
      runner == "rake test"
    end

    def runner
      @runner
    end

    def line_no_arg
      test_name = catch(:test_name) { fetch_test_name }
      return nil if test_name.nil?
      arg = "--name=#{test_name}"
      rake? ? "TESTOPTS=#{arg.inspect}" : arg
    end

    def fetch_test_name
      return nil if line_no.nil?
      current_test_name = nil
      read_file do |current_line_no, line|
        match_test_method(line) { |match| current_test_name = match }
        throw :test_name, current_test_name if current_line_no == line_no
      end
    end

    def match_test_method(line, &block)
      match_active_support line, &block
      match_test_unit line, &block
    end

    def match_active_support(line)
      line.match ACTIVE_SUPPORT_TEST_CASE_MATCHER do |matches|
        quoted_bit = matches[:double_quote] || matches[:single_quote]
        yield "test_#{quoted_bit.gsub(/\s+/, '_')}"
      end
    end

    def match_test_unit(line)
      line.match TEST_UNIT_MATCHER do |matches|
        yield matches[:test_name]
      end
    end

    def read_file
      current_line_no = 0
      File.open file do |io|
        until io.eof?
          current_line_no += 1
          line_text = io.gets
          yield current_line_no, line_text
        end
      end
    end
  end
end
